package com.asperasoft.samples.faspex.cmdclient;

import java.io.*;
import java.net.*;
import java.security.cert.*;
import java.security.KeyManagementException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.net.ssl.*;
import javax.security.auth.x500.X500Principal;

/**
 * A Faspex command-line client.
 * 
 * @author Aspera
 *
 */
abstract class FaspexClient {
    public static final String INBOX_SERVICE_TYPE = "http://schemas.asperasoft.com/srv/pkgaccess";
    public static final String SENDTO_SERVICE_TYPE = "http://schemas.asperasoft.com/srv/pkgsubmit";
    public static final String ISSUE_TOKEN_SERVICE_TYPE = "http://schemas.asperasoft.com/srv/issue-token";
    public static final String DEFAULT_WEB_PATH = "/aspera/faspex";
    
    protected String user;
    protected String password;
    protected String hostPath;
    protected boolean allowInsecureConnection = false;
    
    /**
     * We need to use a custom HostnameVerifier because Java won't handle
     * SSL certificates with IP address as hostname if subjectAltName is not
     * set to the IP address.
     */
    public FaspexClient() {
        HttpsURLConnection.setDefaultHostnameVerifier(new CustomHostnameVerifier());
        try {
            // change the output stream to permit printing of Unicode characters
            System.setOut(new PrintStream(System.out, true, "UTF-8"));
        } catch (UnsupportedEncodingException e) {
            // should never happen
            System.err.println("Unable to handle Unicode.");
            e.printStackTrace();
            System.exit(1);
        }
    }

    /**
     * Run the program: parse the arguments, execute tasks, and handle errors
     * encountered in parsing the arguments and execution.
     * 
     * @param args The command-line arguments to the program.
     */
    public void run(String[] args) {
        try {
            parseArguments(args);
        } catch (Exception e) {
            System.err.println(e.getMessage());
            printUsage();
            System.exit(1);
        }
        try {
            if (allowInsecureConnection)
                disableCertValidation();
            execute();
        } catch (Exception e) {
            System.err.println("Execution failed: " + e.getMessage());
            System.exit(1);
        }
        System.exit(0);
    }
    
    /**
     * Parses the command line arguments passed to the program.
     * 
     * @param args The command line arguments.
     */
    protected abstract void parseArguments(String[] args) throws Exception;
    
    /**
     * Performs the execution of the program
     */
    protected abstract void execute() throws Exception;

    /**
     * Parse login information in the format user:password@host.
     * 
     * @param loginInfo The <code>String</code> to be parsed.
     * @throws IllegalArgumentException If the login info could not be parsed.
     */
    protected void parseLoginInfo(String loginInfo) {
        // match user:password@host
        String patternStr = "^([^:]+):(.+)@([^@]+)$";
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(loginInfo);
        
        boolean matchFound = matcher.find();
        if (!matchFound)
            throw new IllegalArgumentException("Unable to parse login information.");
        
        user = matcher.group(1);
        password = matcher.group(2);
        hostPath = matcher.group(3);
        
        // if no path is specified on the host, default to /aspera/faspex
        if (!hostPath.contains("/"))
            hostPath = hostPath + DEFAULT_WEB_PATH;
    }

    /**
     * Print the usage of this command-line client from the file at the location
     * specified by <code>usageFile</code>.
     * 
     * @param usageFile The location of the usage file to print.
     */
    protected void printUsage() {
        String usageFile = getUsageFile();
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(
                    getClass().getResourceAsStream(usageFile)));
            String line = null;
            while ((line = reader.readLine()) != null)
                System.out.println(line);
        } catch (IOException e) {
            System.err.println("Unable to print usage info.");
        }
    }
    
    /**
     * Get the path of the usage file.
     * 
     * @return The path of the usage file. 
     */
    protected abstract String getUsageFile();
    
    /**
     * Formats an email address for faspex. Faspex has an odd requirement for
     * usernames that everything before the last '@' must be URL escaped.
     * 
     * @param address The address to format
     * @return An email address/username properly formatted for faspex to
     * understand
     */
    protected String urlEscapeEmailAddressForFaspex(String address) {
        String patternStr = "^(.*)@([^@]+)$";
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(address);
        
        boolean matchFound = matcher.find();
        String result;
        if (matchFound) {
            String username = matcher.group(1);
            String domain = matcher.group(2);
            try {
                // URL decode first, in case user has already entered encoded address
                result = URLDecoder.decode(username, "UTF-8");
                result = URLEncoder.encode(result, "UTF-8") + "@" + domain;
                // If your using Faspex 2.0.10.42572 or below uncomment the two lines below, otherwise keep commented out
                //// URLEncoder.encode does not encode plus sign, so do it manually
                //// result = result.replace("+", "%2b");
            } catch (UnsupportedEncodingException e) {
                // won't happen
                throw new RuntimeException("JVM doesn't know UTF-8!?!", e);
            }
        } else {
            result = address;
        }
        
        return result;
    }
    
    /**
     * Finds the URI of the service that is specified by <code>serviceType</code>
     * 
     * @param hostPath The host on which to find the service.
     * @param serviceType The type of the service to be found on the Faspex server.
     * @return The URI of the service.
     * @throws Exception If the service could not be found.
     */
    protected String findService(String hostPath, String serviceType)
             throws ServiceNotFoundException {
        URL url;
        try {
            url = new URL("https://" + hostPath);
        } catch (MalformedURLException e) {
            throw new ServiceNotFoundException("Bad host name or IP address.");
        }
        HttpURLConnection conn;
        try {
            conn = (HttpURLConnection) url.openConnection();
        } catch (IOException e) {
            throw new ServiceNotFoundException("Unable to open HTTP connection to server.");
        }
        // The Accept property must be set to get the server to return an XRDS document.
        conn.setRequestProperty("Accept", "application/xrds+xml");
        
        // Set up a SAX parser to parse the XRDS and find the specified service.
        String serviceUri;
        try {
            ServiceParser parser = new ServiceParser();
            InputStream is = conn.getInputStream();
            serviceUri = parser.findService(is, serviceType);
        } catch (ConnectException e) {
            throw new ServiceNotFoundException("Unable to connect to XRDS service. Please check connection to faspex server.");
        } catch (Exception e) {
            throw new ServiceNotFoundException("Unable to access service of type "
                    + serviceType + ". Please check connection to faspex server.");
        }
        
        return serviceUri;
    }
    
    /**
     * Custom <code>HostnameVerifier</code> that skips the requirement for
     * subjectAltName when the hostname is an IP address.
     * 
     * @author Aspera
     */
    private static class CustomHostnameVerifier implements HostnameVerifier {
        // compare hostname to common name in certificate, not subject
        // alternative name. By default IP addresses get compared to subject alternative name.
        public boolean verify(String hostname, SSLSession session) {
            try {
                X500Principal peerPrincipal = (X500Principal) session.getPeerPrincipal();
                String DN = peerPrincipal.getName("CANONICAL");
                // now parse the CN out of the DN
                String CN = getCN(DN);
                return CN.equals(hostname);
            } catch (SSLPeerUnverifiedException e) {
                return false;
            }
        }
        
        private String getCN(String DN) {
            String[] dnComponents = DN.split(",");
            // Find one that starts with CN=
            for (int i = 0; i <= dnComponents.length; i++) {
                if (dnComponents[i].startsWith("cn=")) {
                    return dnComponents[i].substring(3);
                }
            }
            return "";
        }
    }
    
    /**
     * Disable certificate validation so that connections can be made to faspex
     * servers with self-signed SSL certificates.
     */
    private void disableCertValidation() {
        // Create a trust manager that does not validate certificate chains
        TrustManager[] trustAllCerts = new TrustManager[] {
            new X509TrustManager() {
                public X509Certificate[] getAcceptedIssuers() {
                    return null;
                }
                public void checkClientTrusted(X509Certificate[] certs, String authType) {
                }
                public void checkServerTrusted(X509Certificate[] certs, String authType) {
                }
            }
        };
        
        // Install the all-trusting trust manager
        SSLContext sc;
        try {
            sc = SSLContext.getInstance("SSL");
        sc.init(null, trustAllCerts, new SecureRandom());
        HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());
        } catch (NoSuchAlgorithmException e) {
        } catch (KeyManagementException e) {
        }
    }
}
