package com.asperasoft.samples.faspex.cmdclient;

import java.util.List;

import com.asperasoft.faspmanager.FaspManager;

/**
 * Command-line client for getting a Faspex package from a user's inbox.
 * 
 * @author Aspera
 */
public final class FaspexGet extends FaspexTransferClient {
    private static final String USAGE_FILE = "f_get.txt";
    
    PackageFilter filter = new PackageFilter();
    
    /**
     * The entry point into the command-line FaspexGet program.
     * 
     * @param args The command-line arguments. 
     */
    public static void main(String[] args) {
        FaspexClient program = new FaspexGet();
        program.run(args);
    }
    
    /**
     * Parses the command line arguments passed to the program.
     * 
     * @param args The command line arguments.
     */
    protected void parseArguments(String[] args) throws Exception {
        int i = 0;
        try {
            for (; i < args.length; i++) {
                String arg = args[i];

                if (i == args.length - 1) {
                    // last argument is the local path
                    localPaths = new String[] { arg };
                } else if (i == args.length - 2) {
                    // second to last argument is username:password@hostname
                    parseLoginInfo(arg);
                } else if ("--to".equalsIgnoreCase(arg)) {
                    filter.toFilter = args[++i];
                } else if ("--from".equalsIgnoreCase(arg)) {
                    filter.fromFilter = args[++i];
                } else if ("--subject".equalsIgnoreCase(arg)) {
                    filter.subjectFilter = args[++i];
                } else if ("--message".equalsIgnoreCase(arg)) {
                    filter.messageFilter = args[++i];
                } else if ("--id".equalsIgnoreCase(arg)) {
                    filter.pkgIdFilter = args[++i];
                } else if ("--targetrate".equalsIgnoreCase(arg)) {
                    targetRateKbps = Integer.parseInt(args[++i]);
                } else if ("--retries".equalsIgnoreCase(arg)) {
                    retries = Integer.parseInt(args[++i]);
                } else if ("--url".equalsIgnoreCase(arg)) {
                    userFaspUrl = args[++i];
                } else if ("--insecure".equalsIgnoreCase(arg)) {
                    allowInsecureConnection = true;
                } else {
                    throw new IllegalArgumentException("Unknown option \"" + arg + "\"");
                }
            }

            // check that required values have been populated
            if (user == null || password == null || hostPath == null || localPaths == null)
                throw new IllegalArgumentException("Missing required argument.");
            if (userFaspUrl == null && filter.toFilter == null && filter.fromFilter == null && filter.subjectFilter == null
                    && filter.messageFilter == null && filter.pkgIdFilter == null)
                throw new IllegalArgumentException("At least one filter must be specified.");
        } catch (Exception e) {
            if (i < args.length) {
                String message = "Invalid option or value [\"" + args[i] + "\"] at position [" + i + "]: "
                        + e.getMessage();
                throw new IllegalArgumentException(message);
            } else {
                throw e;
            }
        }
    }
    
    /**
     * Performs the execution of the program.
     */
    protected void execute() throws Exception {
        // if fasp URL is specified, use that, otherwise, find the package from the server
        if (userFaspUrl != null) {
            tryTransfer(userFaspUrl, localPaths, TransferDirection.DOWNLOAD, retries);
        } else {
            // find all packages matching specified filter options and download them
            String inboxServiceURI = findService(hostPath, INBOX_SERVICE_TYPE);
            List<FaspexPackage> pkgList = FaspexList.getPackageList(inboxServiceURI, user, password, filter);
            if (pkgList.size() == 0)
                System.out.println("No packages to download.");
            else {
                for (FaspexPackage pkg : pkgList)
                    downloadPackage(pkg);
            }
        }
        
        FaspManager.getSingleton().destroy();
    }
    
    /**
     * Download the specified package.
     * 
     * @param pkg The package to download.
     * @throws TransferException If the download fails.
     */
    private void downloadPackage(FaspexPackage pkg) throws TransferException {
        System.out.println("Downloading package " + pkg.getTitle()); 
        // first fasp URL is for download of entire package
        tryTransfer(pkg.getFaspUrls().get(0), localPaths, TransferDirection.DOWNLOAD, retries);
    }
    
    /**
     * Get the path of the usage file.
     * 
     * @return The path of the usage file. 
     */
    protected String getUsageFile() {
        return USAGE_FILE;
    }
}
