package com.asperasoft.samples.faspex.cmdclient;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * Command-line client for listing the packages in a user's Faspex inbox.
 * 
 * @author Aspera
 */
public final class FaspexList extends FaspexClient {
    private static final String USAGE_FILE = "f_list.txt";
    
    private PackageFilter filter = new PackageFilter();

    /**
     * The entry point into the command-line FaspexList program.
     * 
     * @param args The command-line arguments. 
     */
    public static void main(String[] args) {
        FaspexClient program = new FaspexList();
        program.run(args);
    }
    
    /**
     * Parses the command line arguments passed to the program.
     * 
     * @param args The command line arguments.
     */
    protected void parseArguments(String[] args) throws Exception {
        int i = 0;
        try {
            for (; i < args.length; i++) {
                String arg = args[i];

                if (i == args.length - 1) {
                    // last argument is username:password@hostname
                    parseLoginInfo(arg);
                } else if ("--to".equalsIgnoreCase(arg)) {
                    filter.toFilter = args[++i];
                } else if ("--from".equalsIgnoreCase(arg)) {
                    filter.fromFilter = args[++i];
                } else if ("--subject".equalsIgnoreCase(arg)) {
                    filter.subjectFilter = args[++i];
                } else if ("--message".equalsIgnoreCase(arg)) {
                    filter.messageFilter = args[++i];
                } else if ("--id".equalsIgnoreCase(arg)) {
                    filter.pkgIdFilter = args[++i];
                } else if ("--insecure".equalsIgnoreCase(arg)) {
                    allowInsecureConnection = true;
                } else {
                    throw new IllegalArgumentException("Unknown option \"" + arg + "\"");
                }
            }

            // check that required values have been populated
            if (user == null || password == null || hostPath == null)
                throw new IllegalArgumentException("Missing required argument.");
        } catch (Exception e) {
            if (i < args.length) {
                String message = "Invalid option or value [\"" + args[i] + "\"] at position [" + i + "]: "
                        + e.getMessage();
                throw new IllegalArgumentException(message);
            } else {
                throw e;
            }
        }
    }
    
    /**
     * Performs the execution of the program.
     */
    protected void execute() throws Exception {
        String inboxServiceURI = findService(hostPath, INBOX_SERVICE_TYPE);
        List<FaspexPackage> pkgList = getPackageList(inboxServiceURI, user, password, filter);
        printPackages(pkgList);
    }
    
    /**
     * Gets the Atom feed for the user's inbox as an <code>InputStream</code>.
     * 
     * @param atomFeedURI The URI for the Atom feed.
     * @param user The user whose inbox will be queried.
     * @param password The user's password.
     * @return The <code>InputStream</code> containing the Atom feed representing
     * the user's inbox.
     * @throws MalformedURLException If the URI of the Atom feed is invalid.
     * @throws IOException If unable to open an HTTP connection to the atom feed.
     */
    private static InputStream getAtomFeed(String atomFeedURI, String user, String password)
            throws MalformedURLException, IOException {
        String authString = Base64Coder.encodeString(user + ":" + password); 
        
        URL url = new URL(atomFeedURI);
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setDoOutput(true);
        conn.setRequestProperty("Authorization", "Basic " + authString);
        
        return conn.getInputStream();
    }
    
    /**
     * Extract the packages from the Atom feed matching the specified filter.
     * 
     * @param atomFeedURI The URI for the Atom feed.
     * @param user The user whose inbox will be queried.
     * @param password The user's password.
     * @param filter The filter to use to determine which packages should be returned.
     * @return A list of packages listed in the Atom feed.
     * @throws IOException If unable to get the Atom feed.
     * @throws PackageListingException If the packages could not be determined from the Atom feed.
     */
    protected static List<FaspexPackage> getPackageList(String atomFeedURI, String user, String password, PackageFilter filter)
            throws IOException, PackageListingException {
        InputStream atomFeed = getAtomFeed(atomFeedURI, user, password);
        PackageParser parser = new PackageParser();
        List<FaspexPackage> packages;
        try {
            packages = parser.getPackageList(atomFeed);
        } catch (Exception e) {
            throw new PackageListingException("Unable to parse packages from Atom feed.");
        }
        
        List<FaspexPackage> filteredPackages = new ArrayList<FaspexPackage>();
        for (FaspexPackage pkg : packages) {
            if (matchesFilters(pkg, filter))
                filteredPackages.add(pkg);
        }
        return filteredPackages;
    }
    
    /**
     * Print the list of packages onto the screen.
     * 
     * @param pkgList The list of packages to print.
     */
    private void printPackages(List<FaspexPackage> pkgList) {
        System.out.println("Packages in " + user + "'s inbox:");
        System.out.println();
        
        if (pkgList.size() == 0)
            System.out.println("No packages found.");
        else {
            for (FaspexPackage pkg : pkgList) {
                printPackage(pkg);
            }
        }
    }
    
    /**
     * Format and print information about a package onto the screen.
     * 
     * @param pkg The package to print.
     */
    private void printPackage(FaspexPackage pkg) {
        System.out.println("ID: " + pkg.getId());
        System.out.println("Subject: " + pkg.getTitle());
        System.out.println("From: " + pkg.getAuthor().getName() 
                + " (" + pkg.getAuthor().getEmail() + ")");
        System.out.println("Files/Directories: ");
        
        // print out the top level files from the package
        List<String> faspURLs = pkg.getFaspUrls();
        for (String faspURL : faspURLs) {
            // file or directory name is between the last '/' and the ? of the fasp URL
            String patternStr = "/([^/\\?]+)\\?";
            Pattern pattern = Pattern.compile(patternStr);
            Matcher matcher = pattern.matcher(faspURL);
            matcher.find();
            String fileOrDirName = matcher.group(1);
            System.out.println("  " + fileOrDirName);
        }
        
        System.out.println();
    }
    
    /**
     * Returns whether the package matches all the search filters.
     * 
     * @param pkg The package to compare against the filters.
     * @return true if the package matches all filters, false otherwise.
     */
    private static boolean matchesFilters(FaspexPackage pkg, PackageFilter filter) {
        String email;
        String name;
        
        if (filter.fromFilter != null) {
            email = pkg.getAuthor().getEmail().toLowerCase();
            name = pkg.getAuthor().getName().toLowerCase();
            if (!matchesFilter(filter.fromFilter.toLowerCase(), email)
                    && !matchesFilter(filter.fromFilter.toLowerCase(), name))
                return false;
        }
        if (filter.toFilter != null) {
            List<FaspexUser> recipients = pkg.getRecipients();
            boolean matchFound = false;
            for (FaspexUser recipient : recipients) {
                email = recipient.getEmail().toLowerCase();
                name = recipient.getName().toLowerCase();
                if (!matchesFilter(filter.toFilter.toLowerCase(), email)
                        && !matchesFilter(filter.toFilter.toLowerCase(), name))
                    matchFound = true;
            }
            if (!matchFound)
                return false;
        }
        if (filter.subjectFilter != null) {
            String subject = pkg.getTitle().toLowerCase();
            if (!matchesFilter(filter.subjectFilter.toLowerCase(), subject))
                return false;
        }
        if (filter.messageFilter != null) {
            String message = pkg.getContent().toLowerCase();
            if (!matchesFilter(filter.messageFilter.toLowerCase(), message))
                return false;
        }
        if (filter.pkgIdFilter != null) {
            String pkgId = pkg.getId().toLowerCase();
            if (!matchesFilter(filter.pkgIdFilter.toLowerCase(), pkgId))
                return false;
        }

        return true;
    }
    
    /**
     * Returns whether or not the given text matches a glob filter
     * 
     * @param filter The filter to match the text with.
     * @param text The text to match with the filter.
     * @return Whether or not the text matches the filter.
     */
    private static boolean matchesFilter(String filter, String text) {
        String regex = convertGlobToRegEx(filter);
        Pattern pattern = Pattern.compile(regex);
        Matcher matcher = pattern.matcher(text);
        return matcher.matches();
    }
    
    /**
     * Converts a glob expression into a regular expression
     * 
     * @param globExpression The glob expression to convert into a regular expression.
     * @return The converted regular expression.
     */
    private static String convertGlobToRegEx(String globExpression) {
        globExpression = globExpression.trim();
        int strLen = globExpression.length();
        StringBuilder sb = new StringBuilder(strLen);
        boolean escaping = false;
        int inCurlies = 0;
        for (char currentChar : globExpression.toCharArray()) {
            switch (currentChar) {
            case '*':
                if (escaping)
                    sb.append("\\*");
                else
                    sb.append(".*");
                escaping = false;
                break;
            case '?':
                if (escaping)
                    sb.append("\\?");
                else
                    sb.append('.');
                escaping = false;
                break;
            case '.':
            case '(':
            case ')':
            case '+':
            case '|':
            case '^':
            case '$':
            case '@':
            case '%':
                sb.append('\\');
                sb.append(currentChar);
                escaping = false;
                break;
            case '\\':
                if (escaping) {
                    sb.append("\\\\");
                    escaping = false;
                } else
                    escaping = true;
                break;
            case '{':
                if (escaping) {
                    sb.append("\\{");
                } else {
                    sb.append('(');
                    inCurlies++;
                }
                escaping = false;
                break;
            case '}':
                if (inCurlies > 0 && !escaping) {
                    sb.append(')');
                    inCurlies--;
                } else if (escaping)
                    sb.append("\\}");
                else
                    sb.append("}");
                escaping = false;
                break;
            case ',':
                if (inCurlies > 0 && !escaping) {
                    sb.append('|');
                } else if (escaping)
                    sb.append("\\,");
                else
                    sb.append(",");
                break;
            default:
                escaping = false;
                sb.append(currentChar);
            }
        }
        return sb.toString();
}
    
    /**
     * Get the path of the usage file.
     * 
     * @return The path of the usage file. 
     */
    protected String getUsageFile() {
        return USAGE_FILE;
    }
}
