package com.asperasoft.samples.faspex.cmdclient;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.asperasoft.faspmanager.FaspManager;
import org.apache.commons.lang.StringEscapeUtils;

/**
 * Command-line client for sending a Faspex package.
 * 
 * @author Aspera
 */
public final class FaspexSend extends FaspexTransferClient {
    private static final String USAGE_FILE = "f_send.txt";
    private static final int HTTP_RESPONSE_OK = 200; // HTTP response code for OK
    
    private String[] recipients;
    private String from;
    private String subject;
    private String message;

    /**
     * The entry point into the command-line FaspexSend program.
     * 
     * @param args The command-line arguments. 
     */
    public static void main(String[] args) {

        FaspexClient program = new FaspexSend();
        program.run(args);
    }
    
    /**
     * Parses the command line arguments passed to the program.
     * 
     * @param args The command line arguments.
     */
    protected void parseArguments(String[] args) throws Exception {
        int i = 0;
        try {
            for (; i < args.length; i++) {
                String arg = args[i];

                if (i == args.length - 1) {
                    // last argument is the local path
                    localPaths = arg.split(",");
                } else if (i == args.length - 2) {
                    // second to last argument is username:password@hostname
                    parseLoginInfo(arg);
                } else if ("--to".equalsIgnoreCase(arg)) {
                    recipients = args[++i].split(",");
                } else if ("--from".equalsIgnoreCase(arg)) {
                    from = args[++i];
                } else if ("--subject".equalsIgnoreCase(arg)) {
                    subject = args[++i];
                } else if ("--message".equalsIgnoreCase(arg)) {
                    message = args[++i];
                } else if ("--targetrate".equalsIgnoreCase(arg)) {
                    targetRateKbps = Integer.parseInt(args[++i]);
                } else if ("--retries".equalsIgnoreCase(arg)) {
                    retries = Integer.parseInt(args[++i]);
                } else if ("--url".equalsIgnoreCase(arg)) {
                    userFaspUrl = args[++i];
                } else if ("--insecure".equalsIgnoreCase(arg)) {
                    allowInsecureConnection = true;
                } else {
                    throw new IllegalArgumentException("Unknown option \"" + arg + "\"");
                }
            }

            // check that required values have been populated
            if ((user == null || password == null || hostPath == null || localPaths == null)
                    || userFaspUrl == null && (recipients == null || subject == null))
                throw new IllegalArgumentException("Missing required argument.");
        } catch (Exception e) {
            if (i < args.length) {
                String message = "Invalid option or value [\"" + args[i] + "\"] at position [" + i + "]: "
                        + e.getMessage();
                throw new IllegalArgumentException(message);
            } else {
                throw e;
            }
        }
    }

    /**
     * Performs the execution of the program.
     */
    protected void execute() throws Exception {
        
        // If fasp URL was specified, use that. Otherwise, get it from the server.
        String faspUrl;
        if (userFaspUrl != null)
            faspUrl = userFaspUrl;
        else
            faspUrl = getFaspURL();
        
        tryTransfer(faspUrl, localPaths, TransferDirection.UPLOAD, retries);
        
        FaspManager.getSingleton().destroy();
    }
    
    /**
     * Gets the fasp URL to use for the transfer from the send-to service.
     * 
     * @return The fasp URL
     * @throws Exception If unable to get the fasp URL.
     */
    private String getFaspURL() throws Exception {
        String sendToServiceUri = findService(hostPath, SENDTO_SERVICE_TYPE);
        // HTTP basic authorization uses base 64 encoding of the username and password
        String authString = Base64Coder.encodeString(user + ":" + password);

        // use the send-to service to get a fasp URL for transfer
        URL url = new URL(sendToServiceUri);
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setRequestMethod("POST");
        conn.setDoOutput(true);
        conn.setRequestProperty("Authorization", "Basic " + authString);

        StringBuilder toBuilder = new StringBuilder();
        for (String recipient : recipients) {
            // a workaround for faspex expecting all characters before the last '@'
            // to be URL escaped.
            recipient = urlEscapeEmailAddressForFaspex(recipient);
            toBuilder.append("<to>" + StringEscapeUtils.escapeXml(recipient) + "</to>");
        }
        String escapedTo = toBuilder.toString();
        String escapedFrom = from == null ? "" : StringEscapeUtils.escapeXml(urlEscapeEmailAddressForFaspex(from));
        
        String httpBody =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
            "<package xmlns=\"http://schemas.asperasoft.com/xml/package\">" +
                escapedTo +
                ((from == null) ? "" : "<from>" + escapedFrom + "</from>") +
                "<subject>" + StringEscapeUtils.escapeXml(subject) + "</subject>" +
                ((message == null) ? "" : "<body>" + StringEscapeUtils.escapeXml(message) + "</body>") +
            "</package>";
        PrintWriter out = new PrintWriter(conn.getOutputStream());
        out.println(httpBody);
        out.flush();
        
        // read the HTTP response and parse out the fasp URL
        if (conn.getResponseCode() != HTTP_RESPONSE_OK)
            throw new FaspexClientException("Unable to send to specified user: " + conn.getResponseMessage());
        BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        String line;
        StringBuilder sb = new StringBuilder();
        while ((line = in.readLine()) != null) {
            sb.append(line);
            sb.append(System.getProperty("line.separator"));
        }
        String response = sb.toString();

        String patternStr = "\"(fasp://[^\"]*)\"";
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(response);
        
        boolean matchFound = matcher.find();
        if (!matchFound)
            throw new FaspexClientException("Unable to find valid fasp URL in response.");
        
        // double unescape the URL because it may be double escaped
        String faspURL = StringEscapeUtils.unescapeXml(StringEscapeUtils.unescapeXml(matcher.group(1)));
        return faspURL;
    }
    
    /**
     * Get the path of the usage file.
     * 
     * @return The path of the usage file. 
     */
    protected String getUsageFile() {
        return USAGE_FILE;
    }
}
