package com.asperasoft.samples.faspex.cmdclient;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.commons.lang.StringEscapeUtils;


abstract class FaspexTransferClient extends FaspexClient {
    protected static final int ERR_UNAUTHORIZED = 34; // error code for token failure
    
    protected String[] localPaths;
    protected int targetRateKbps = 0;
    protected int retries = 3; // default to 3 retries
    protected String userFaspUrl; // user defined URL entered on command line

    /**
     * Create a transfer job given the specified parameters. Try to perform a
     * transfer up to the specified number of retries, performing token refresh
     * as necessary.
     * 
     * @param job The job specifying the transfer to perform.
     * @param retries The number of retries to perform in case of failure.
     * @throws TransferException If the number of retries is exceeded.
     */
    protected void tryTransfer(String faspUrl, String[] localPaths, TransferDirection direction, int retries) throws TransferException {
        // Obtain a token for the transfer and append it to the fasp URL for downloads
        if (direction == TransferDirection.DOWNLOAD) {
            String token = null;
            try {
                token = getToken(faspUrl);
            } catch (Exception e) {
                throw new TransferException("Unable to get token for transfer.");
            }
            faspUrl = faspUrl + "&token=" + token;
        }

        TransferJob job = newJob(faspUrl, localPaths, direction);
        boolean done = false;
        int triesRemaining = retries + 1; // include the first (non-retry) try
        while (!done && triesRemaining > 0) {
            job.startTransfer();
            job.monitorTransfer();
            triesRemaining--;
            if (job.getErrorCode() == 0) {
                done = true;
            } else {
                if (triesRemaining > 0) {
                    System.out.println("Retrying transfer.");
                } else {
                    throw new TransferException("Exceeded number of retries.");
                }
            }
        }
    }
    
    private TransferJob newJob(String faspUrl, String[] localPaths, TransferDirection direction) throws TransferException {
        TransferJob job;
        try {
            job = new TransferJob(faspUrl, localPaths, direction);
        } catch (MalformedURLException e) {
            throw new TransferException("Invalid fasp URL: " + e.getMessage());
        }
        if (targetRateKbps > 0 )
            job.setTargetRate(targetRateKbps);
        return job;
    }
    
    private String getToken(String faspUrl) throws Exception {
        String serviceUri = findService(hostPath, ISSUE_TOKEN_SERVICE_TYPE);
        
        // HTTP basic authorization uses base 64 encoding of the username and password
        String authString = Base64Coder.encodeString(user + ":" + password);

        // use the send-to service to get a fasp URL for transfer
        URL url = new URL(serviceUri);
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setRequestMethod("POST");
        conn.setDoOutput(true);
        conn.setRequestProperty("Authorization", "Basic " + authString);
        
        String httpBody =
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" +
            "<url-list xmlns=\"http://schemas.asperasoft.com/xml/url-list\">" +
                 "<url href=\"" + StringEscapeUtils.escapeXml(faspUrl) + "\"/>" +
            "</url-list>";
        PrintWriter out = new PrintWriter(conn.getOutputStream());
        out.println(httpBody);
        out.flush();
        
        // read the HTTP response and parse out the fasp URL
        BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        String token = in.readLine();
        in.close();
        return token;
    }
}
