package com.asperasoft.samples.faspex.cmdclient;

import java.net.MalformedURLException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.asperasoft.faspmanager.*;

/**
 * A transfer job that can be used to perform transfers after the connection
 * and transfer parameters have been set. The environment variable ASCP_PATH
 * must be set to point to the location of the ascp binary.
 * 
 * @author Aspera
 */
public class TransferJob implements TransferListener {
    private static final String ASCP_PATH = "ascp";
    private static final String PRIVATE_KEY_LOCATION = "asperaweb_id_dsa.putty";
    private static final int MONITOR_INTERVAL = 2000; // milliseconds between refreshing statistics
    
    // connection related parameters
    private String user;
    private String password;
    private String host;
    private String[] localPaths;
    private String remotePath;
    private XferParams xferParams = new XferParams();
    private String sessionId; // session ID of the transfer associated with this connection.
    private boolean lockTargetRate = false; // target rate cannot be changed if locked.
    private TransferDirection direction;
    
    // current state of the transfer
    private boolean finished = false;
    private int errorCode = 0;
    private String errorDescription;
    
    static {
        Environment.setFasp2ScpPath(ASCP_PATH);
        Environment.setUnicodeSupportOn(true);
    }
    
    /**
     * Constructs a <code>TransferJob</code> using the specified fasp URLs and
     * local paths. Only the first fasp URL is used to set the connection
     * properties. The rest are assumed to be the same except for the remote
     * path.
     * 
     * @param faspURL The fasp URL to use to create a connection.
     * @param localPaths The local paths to the files to be transferred or the
     * destination for downloads.
     * @throws MalformedURLException If the fasp URL cannot be parsed.
     * @throws TransferException If the ASCP_PATH environment variable has not been set.
     */
    public TransferJob(String faspURL, String[] localPaths, TransferDirection direction)
            throws MalformedURLException, TransferException {
        if (ASCP_PATH == null)
            throw new TransferException("Please set ASCP_PATH environment variable.");
        
        this.localPaths = localPaths;
        this.direction = direction;
        
        // fasp URL is in the format fasp://<user>:<password>@<host>:<port>/<path>?<parameters>
        String patternStr = "^fasp://(([^@:/]+)(:([^@:/]+))?@)?([^@/:]+)(:([0-9]+))?(/[^\\?]+)\\?(.*)$";
        Pattern pattern = Pattern.compile(patternStr);
        Matcher matcher = pattern.matcher(faspURL);
        
        boolean matchFound = matcher.find();
        if (!matchFound)
            throw new MalformedURLException("Unable to parse fasp URL.");
        
        user = matcher.group(2);
        password = matcher.group(4);
        host = matcher.group(5);
        if (matcher.group(7) != null) {
            int tcpPort = Integer.parseInt(matcher.group(7));
            xferParams.tcpPort = tcpPort;
        }
        remotePath = matcher.group(8);
        
        // parameter names and values alternate, split by "&amp;" and "="
        String paramStr = matcher.group(9);
        String[] params = paramStr.split("&amp;|&|=");
        for (int i = 0; i < params.length; i = i + 2) {
            String param = params[i];
            String value = params[i + 1];
            setXferParam(param, value);
        }
        
        xferParams.preCalculateJobSize = true; // so we know the size of the transfer beforehand
    }
    
    /**
     * Starts a transfer to/from the faspex server.
     * 
     * @return The job ID of the transfer.
     * @throws TransferException If the transfer fails.
     */
    public String startTransfer() throws TransferException {
        // reset the transfer state
        finished = false;
        errorCode = 0;
        errorDescription = null;
        
       // Create a local source location object
        LocalLocation local = new LocalLocation();
        for (String localPath : localPaths) {
            local.addPath(localPath);
        }
        
        // Create a remote destination object
        RemoteLocation remote;
        if (password != null)
            // use password if one is specified 
            remote = new RemoteLocation(host, user, password);
        else
            // otherwise, use public-key authentication
            remote = new RemoteLocation(host, user, PRIVATE_KEY_LOCATION, null);
        
        // Add remote path to the destination object
        remote.addPath(remotePath);
        
        // Setup a job order. A job order consists of source, destination and
        // the transfer params (XferParams object)
        TransferOrder order;
        
        if (direction == TransferDirection.DOWNLOAD)
            order = new TransferOrder(remote, local, xferParams);
        else
            order = new TransferOrder(local, remote, xferParams);
        
        if (direction == TransferDirection.UPLOAD) {
            System.out.println("Starting upload.");
        } else {
            // find the name of the file or directory to be downloaded
            String patternStr = "/([^/]+)$";
            Pattern pattern = Pattern.compile(patternStr);
            Matcher matcher = pattern.matcher(remotePath);
            matcher.find();
            String fileOrDirName = matcher.group(1);
            System.out.println("Starting download of " + fileOrDirName);
        }
        
        // Submit the transfer order
        try {
            sessionId = FaspManager.getSingleton().startTransfer(order, this);
        } catch (Exception e) {
            throw new TransferException("Unable to find or start ascp.");
        }
        
        return sessionId;
    }
    
    /**
     * Blocking call to monitor the status of the transfer and print status updates to the screen.
     * 
     * @throws TransferException If an error occurs during the transfer.
     */
    public void monitorTransfer() {
        while (!finished) {
            try {
                SessionStats stats = FaspManager.getSingleton().getSessionStats(sessionId);
                SessionState state = stats.getState();
                
                if (state == SessionState.STARTING || state == SessionState.CONNECTING) {
                    // simply wait for the transfer to start
                    Thread.sleep(MONITOR_INTERVAL);
                } else if (state == SessionState.TRANSFERRING) {
                    // transferring, write out statistics
                    long completedBytes = stats.getTotalWrittenBytes();
                    long totalBytes = stats.getPreCalcTotalBytes();
                    long percent = 0;
                    if (totalBytes > 0)
                        percent = completedBytes * 100 / totalBytes;
                    
                    // overwrite previous transfer status line with current line
                    String output = "\r" + percent + " percent complete: "
                            + formatBytes(completedBytes) + " of " 
                            + formatBytes(totalBytes) + " transferred.";
                    System.out.printf("%-80.80s", output);
                    
                    Thread.sleep(MONITOR_INTERVAL);
                }
            } catch (SessionNotFoundException e) {
                finished = true;
            } catch (InterruptedException e) {
                // just ignore and go on
            } catch (InitializationException e) {
                //ignore
            }
        }
        
        // overwrite transfer status with "Transfer completed" or "Transfer failed"
        if (errorCode == 0)
            System.out.printf("%-80s", "\rTransfer completed.");
        else
            System.out.printf("%-80s", "\rTransfer failed: " + errorDescription);
        System.out.println();
    }
    
    /**
     * Sets the specified transfer parameter to the specified value.
     * 
     * @param param The parameter to set.
     * @param value The value to set the parameter to.
     */
    private void setXferParam(String param, String value) {
        if (param.equalsIgnoreCase("token")) {
            xferParams.token = value;
        } else if (param.equalsIgnoreCase("cookie")) {
            xferParams.cookie = value;
        } else if (param.equalsIgnoreCase("v")) {
            if (value.equals("2"))
                xferParams.faspVersion = Version.FASP2;
            else
                xferParams.faspVersion = Version.FASP1;
        } else if (param.equalsIgnoreCase("port")) {
            xferParams.udpPort = Integer.parseInt(value);
        } else if (param.equalsIgnoreCase("pathmtu")) {
            if (value.equalsIgnoreCase("auto"))
                xferParams.datagramSize = 0;
            else
                xferParams.datagramSize = Integer.parseInt(value);
        } else if (param.equalsIgnoreCase("policy")) {
            if (value.equalsIgnoreCase("fixed"))
                xferParams.policy = Policy.FIXED;
            else if (value.equalsIgnoreCase("trickle"))
                xferParams.policy = Policy.LOW;
            else
                xferParams.policy = Policy.FAIR;
        } else if (param.equalsIgnoreCase("linkcapacity")) {
            if (value.equalsIgnoreCase("auto"))
                xferParams.autoDetectCapacity = true;
            else {
                xferParams.autoDetectCapacity = false;
            }
        } else if (param.equalsIgnoreCase("targetrate")) {
            xferParams.targetRateKbps = Integer.parseInt(value);
        } else if (param.equalsIgnoreCase("minrate")) {
            xferParams.minimumRateKbps = Integer.parseInt(value);
        } else if (param.equalsIgnoreCase("locktargetrate")) {
            if (value.equalsIgnoreCase("yes"))
                lockTargetRate = true;
            else
                lockTargetRate = false;
        } else if (param.equalsIgnoreCase("enc")) {
            if (value.equalsIgnoreCase("aes128"))
                xferParams.encryption = Encryption.AES128;
            else
                xferParams.encryption = Encryption.NONE;
        } else if (param.equalsIgnoreCase("createpath")) {
            if (value.equalsIgnoreCase("yes"))
                xferParams.createPath = true;
            else
                xferParams.createPath = false;
        } else {
            // either unknown parameter or valid parameter that is not handled by fasp manager
            // do nothing
        }
    }
    
    /**
     * Sets the target rate for the transfer. Does nothing if target rate is locked.
     * 
     * @param targetRateKbps The new target rate for the transfer.
     */
    public void setTargetRate(int targetRateKbps) {
        if (targetRateKbps > 0 && !lockTargetRate) {
            xferParams.targetRateKbps = targetRateKbps;
        }
    }
    
    /**
     * Get the error code of the ended session.
     * 
     * @return The error code or 0 if there was no error.
     */
    public int getErrorCode() {
        return errorCode;
    }
    
    /**
     * Get the error description of the ended session.
     * 
     * @return The error description or null if there was no error.
     */
    public String getErrorDescription() {
        return errorDescription;
    }
    
    /**
     * Gets whether or not the transfer has finished.
     * 
     * @return <code>true</code> if the transfer has finished, <code>false</code> otherwise.
     */
    public boolean isFinished() {
        return finished;
    }
   
    /**
     * Format the number of bytes into a more readable form. E.g. 1024 bytes
     * becomes 1 KB etc.
     * 
     * @param bytes The input to be formatted.
     * @return A formatted representation of the number of bytes.
     */
    private String formatBytes(long bytes) {
        final int kilobyte = 1024;
        final int megabyte = 1024 * 1024;
        final int gigabyte = 1024 * 1024 * 1024;
        
        String result;
        float size;
        
        if (bytes < kilobyte) {
            result = bytes + " bytes";
        } else if (bytes < megabyte) {
            size = (float) bytes / kilobyte;
            result = String.format("%.2f", size) + " KB";
        } else if (bytes < gigabyte) {
            size = (float) bytes / megabyte;
            result = String.format("%.2f", size) + " MB";
        } else {
            size = (float) bytes / gigabyte;
            result = String.format("%.2f", size) + " GB";
        }
            
        return result;            
    }
    
    /**
     * Listener for session events.
     * 
     * @param event The event that caused the listener to be called.
     * @param sessionStats The current statistics of the session.
     * @param fileInfo
     */
    public void fileSessionEvent(TransferEvent event, SessionStats sessionStats, FileInfo fileInfo) {
        if (event == TransferEvent.SESSION_STOP || event == TransferEvent.SESSION_ERROR)
            finished = true;
        if (event == TransferEvent.SESSION_ERROR) {
            errorCode = sessionStats.getErrorCode();
            errorDescription = sessionStats.getErrorDescription();
        }
    }
}
